@@ -32,7 +32,7 @@
 
 #ifndef lint
 static char ocopyright[] =
-"$Id: dhclient.c,v 1.129.2.36 2007/05/01 20:42:55 each Exp $ Copyright (c) 2004-2006 Internet Systems Consortium.  All rights reserved.\n";
+"$Id: dhclient.c,v 1.143.2.6 2007/05/22 20:37:04 each Exp $ Copyright (c) 2004-2007 Internet Systems Consortium.  All rights reserved.\n";
 #endif /* not lint */
 
 #include "dhcpd.h"
@@ -77,12 +77,11 @@ int nowait=0;
 
 static void usage PROTO ((void));
 
-void do_release(struct client_state *);
-
 int main (argc, argv, envp)
 	int argc;
 	char **argv, **envp;
 {
+	int fd;
 	int i;
 	struct servent *ent;
 	struct interface_info *ip;
@@ -91,6 +90,7 @@ int main (argc, argv, envp)
 	char *server = (char *)0;
 	char *relay = (char *)0;
 	isc_result_t status;
+ 	int exit_mode = 0;
  	int release_mode = 0;
 	omapi_object_t *listener;
 	isc_result_t result;
@@ -102,15 +102,18 @@ int main (argc, argv, envp)
 	int no_dhclient_script = 0;
 	char *s;
 
-	/* Make sure we have stdin, stdout and stderr. */
-	i = open ("/dev/null", O_RDWR);
-	if (i == 0)
-		i = open ("/dev/null", O_RDWR);
-	if (i == 1) {
-		i = open ("/dev/null", O_RDWR);
-		log_perror = 0; /* No sense logging to /dev/null. */
-	} else if (i != -1)
-		close (i);
+        /* Make sure that file descriptors 0 (stdin), 1, (stdout), and
+           2 (stderr) are open. To do this, we assume that when we
+           open a file the lowest available file decriptor is used. */
+        fd = open("/dev/null", O_RDWR);
+        if (fd == 0)
+                fd = open("/dev/null", O_RDWR);
+        if (fd == 1)
+                fd = open("/dev/null", O_RDWR);
+        if (fd == 2)
+                log_perror = 0; /* No sense logging to /dev/null. */
+        else if (fd != -1)
+                close(fd);
 
 #ifdef SYSLOG_4_2
 	openlog ("dhclient", LOG_NDELAY);
@@ -141,6 +144,10 @@ int main (argc, argv, envp)
 		if (!strcmp (argv [i], "-r")) {
 			release_mode = 1;
 			no_daemon = 1;
+                } else if (!strcmp (argv [i], "-x")) { /* eXit, no release */
+                        release_mode = 0;
+                        no_daemon = 0;
+                        exit_mode = 1;
 		} else if (!strcmp (argv [i], "-p")) {
 			if (++i == argc)
 				usage ();
@@ -241,7 +248,7 @@ int main (argc, argv, envp)
 	}
 
 	/* first kill of any currently running client */
-	if (release_mode) {
+	if (release_mode || exit_mode) {
 		FILE *pidfd;
 		pid_t oldpid;
 		long temp;
@@ -365,7 +372,7 @@ int main (argc, argv, envp)
 			log_info ("No broadcast interfaces found - exiting.");
 			exit (0);
 		}
-	} else if (!release_mode) {
+	} else if (!release_mode && !exit_mode) {
 		/* Call the script with the list of interfaces. */
 		for (ip = interfaces; ip; ip = ip -> next) {
 			/* If interfaces were specified, don't configure
@@ -411,7 +418,9 @@ int main (argc, argv, envp)
 	for (ip = interfaces; ip; ip = ip -> next) {
 		ip -> flags |= INTERFACE_RUNNING;
 		for (client = ip -> client; client; client = client -> next) {
-			if (release_mode)
+                        if (exit_mode)
+                                state_stop (client);
+                        else if (release_mode)
 				do_release (client);
 			else {
 				client -> state = S_INIT;
@@ -423,7 +432,7 @@ int main (argc, argv, envp)
 		}
 	}
 
-	if (release_mode)
+	if (release_mode || exit_mode)
 		return 0;
 
 	/* Start up a listener for the object management API protocol. */
@@ -476,7 +485,7 @@ static void usage ()
 	log_info (arr);
 	log_info (url);
 
-	log_error ("Usage: dhclient [-1dqr] [-nw] [-p <port>] %s",
+	log_error ("Usage: dhclient [-1dqrx] [-nw] [-p <port>] %s",
 		   "[-s server]");
 	log_error ("                [-cf config-file] [-lf lease-file]%s",
 		   "[-pf pid-file] [-e VAR=val]");
@@ -998,7 +1007,9 @@ void db_startup (testp)
 void bootp (packet)
 	struct packet *packet;
 {
-	struct iaddrlist *ap;
+	struct iaddrmatchlist *ap;
+	char addrbuf[4*16];
+	char maskbuf[4*16];
 
 	if (packet -> raw -> op != BOOTREPLY)
 		return;
@@ -1007,9 +1018,17 @@ void bootp (packet)
 	   on it. */
 	for (ap = packet -> interface -> client -> config -> reject_list;
 	     ap; ap = ap -> next) {
-		if (addr_eq (packet -> client_addr, ap -> addr)) {
-			log_info ("BOOTREPLY from %s rejected.",
-			      piaddr (ap -> addr));
+		if (addr_match(&packet->client_addr, &ap->match)) {
+
+		        /* piaddr() returns its result in a static
+			   buffer sized 4*16 (see common/inet.c). */
+
+		        strcpy(addrbuf, piaddr(ap->match.addr));
+		        strcpy(maskbuf, piaddr(ap->match.mask));
+
+			log_info("BOOTREPLY from %s rejected by rule %s "
+				 "mask %s.", piaddr(packet->client_addr),
+				 addrbuf, maskbuf);
 			return;
 		}
 	}
@@ -1021,9 +1040,11 @@ void bootp (packet)
 void dhcp (packet)
 	struct packet *packet;
 {
-	struct iaddrlist *ap;
+	struct iaddrmatchlist *ap;
 	void (*handler) PROTO ((struct packet *));
 	const char *type;
+	char addrbuf[4*16];
+	char maskbuf[4*16];
 
 	switch (packet -> packet_type) {
 	      case DHCPOFFER:
@@ -1049,9 +1070,17 @@ void dhcp (packet)
 	   on it. */
 	for (ap = packet -> interface -> client -> config -> reject_list;
 	     ap; ap = ap -> next) {
-		if (addr_eq (packet -> client_addr, ap -> addr)) {
-			log_info ("%s from %s rejected.",
-			      type, piaddr (ap -> addr));
+		if (addr_match(&packet->client_addr, &ap->match)) {
+
+		        /* piaddr() returns its result in a static
+			   buffer sized 4*16 (see common/inet.c). */
+
+		        strcpy(addrbuf, piaddr(ap->match.addr));
+		        strcpy(maskbuf, piaddr(ap->match.mask));
+
+			log_info("%s from %s rejected by rule %s mask %s.",
+				 type, piaddr(packet->client_addr),
+				 addrbuf, maskbuf);
 			return;
 		}
 	}
@@ -1104,13 +1133,22 @@ void dhcpoffer (packet)
 		if (!lookup_option
 		    (&dhcp_universe, packet -> options,
 		     client -> config -> required_options [i])) {
-		    log_info ("%s: no %s option.",
-			      obuf, (dhcp_universe.options
-				     [client -> config -> required_options [i]]
-				     -> name));
-				return;
-			}
+		    struct option *option = NULL;
+		    unsigned code = client->config->required_options[i];
+
+		    option_code_hash_lookup(&option, dhcp_universe.code_hash,
+					    &code, 0, MDL);
+
+		    if (option)
+			log_info("%s: no %s option.", obuf, option->name);
+		    else
+			log_info("%s: no unknown-%u option.", obuf, code);
+
+		    option_dereference(&option, MDL);
+
+		    return;
 		}
+	    }
 	}
 
 	/* If we've already seen this lease, don't record it again. */
@@ -1184,6 +1222,7 @@ struct client_lease *packet_to_lease (packet, client)
 	struct client_lease *lease;
 	unsigned i;
 	struct option_cache *oc;
+	struct option *option = NULL;
 	struct data_string data;
 
 	lease = (struct client_lease *)new_client_lease (MDL);
@@ -1216,11 +1255,18 @@ struct client_lease *packet_to_lease (packet, client)
 					   packet -> options, lease -> options,
 					   &global_scope, oc, MDL)) {
 			if (data.len) {
+				if (!option_code_hash_lookup(&option,
+						dhcp_universe.code_hash,
+						&i, 0, MDL))
+					log_fatal("Unable to find VENDOR "
+						  "option (%s:%d).", MDL);
 				parse_encapsulated_suboptions
-					(packet -> options, &dhcp_options [i],
+					(packet -> options, option,
 					 data.data, data.len, &dhcp_universe,
 					 client -> config -> vendor_space_name
 						);
+
+				option_dereference(&option, MDL);
 			}
 			data_string_forget (&data, MDL);
 		}
@@ -1810,6 +1856,7 @@ void make_client_options (client, lease, type, sid, rip, prl, op)
 {
 	unsigned i;
 	struct option_cache *oc;
+	struct option *option = NULL;
 	struct buffer *bp = (struct buffer *)0;
 
 	/* If there are any leftover options, get rid of them. */
@@ -1828,26 +1875,31 @@ void make_client_options (client, lease, type, sid, rip, prl, op)
 	/* Send the requested address if provided. */
 	if (rip) {
 		client -> requested_address = *rip;
-		if (!(make_const_option_cache
-		      (&oc, (struct buffer **)0, rip -> iabuf, rip -> len,
-		       &dhcp_options [DHO_DHCP_REQUESTED_ADDRESS], MDL)))
+		i = DHO_DHCP_REQUESTED_ADDRESS;
+		if (!(option_code_hash_lookup(&option, dhcp_universe.code_hash,
+					      &i, 0, MDL) &&
+		      make_const_option_cache(&oc, NULL, rip->iabuf, rip->len,
+					      option, MDL)))
 			log_error ("can't make requested address cache.");
 		else {
 			save_option (&dhcp_universe, *op, oc);
 			option_cache_dereference (&oc, MDL);
 		}
+		option_dereference(&option, MDL);
 	} else {
 		client -> requested_address.len = 0;
 	}
 
-	if (!(make_const_option_cache
-	      (&oc, (struct buffer **)0,
-	       type, 1, &dhcp_options [DHO_DHCP_MESSAGE_TYPE], MDL)))
+	i = DHO_DHCP_MESSAGE_TYPE;
+	if (!(option_code_hash_lookup(&option, dhcp_universe.code_hash, &i, 0,
+				      MDL) &&
+	      make_const_option_cache(&oc, NULL, type, 1, option, MDL)))
 		log_error ("can't make message type.");
 	else {
 		save_option (&dhcp_universe, *op, oc);
 		option_cache_dereference (&oc, MDL);
 	}
+	option_dereference(&option, MDL);
 
 	if (prl) {
 		/* Figure out how many parameters were requested. */
@@ -1856,17 +1908,21 @@ void make_client_options (client, lease, type, sid, rip, prl, op)
 		if (!buffer_allocate (&bp, i, MDL))
 			log_error ("can't make parameter list buffer.");
 		else {
+			unsigned code = DHO_DHCP_PARAMETER_REQUEST_LIST;
+
 			for (i = 0; prl [i]; i++)
 				bp -> data [i] = prl [i];
-			if (!(make_const_option_cache
-			      (&oc, &bp, (u_int8_t *)0, i,
-			       &dhcp_options [DHO_DHCP_PARAMETER_REQUEST_LIST],
-			       MDL)))
+			if (!(option_code_hash_lookup(&option,
+						      dhcp_universe.code_hash,
+						      &code, 0, MDL) &&
+			      make_const_option_cache(&oc, &bp, NULL, i,
+						      option, MDL)))
 				log_error ("can't make option cache");
 			else {
 				save_option (&dhcp_universe, *op, oc);
 				option_cache_dereference (&oc, MDL);
 			}
+			option_dereference(&option, MDL);
 		}
 	}
 
@@ -2254,6 +2310,7 @@ int write_client_lease (client, lease, rewrite, makesure)
 	pair *hash;
 	int errors = 0;
 	char *s;
+	const char *tval;
 
 	if (!rewrite) {
 		if (leases_written++ > 20) {
@@ -2316,15 +2373,15 @@ int write_client_lease (client, lease, rewrite, makesure)
 			errors++;
 
 	}
-	if (lease -> server_name) {
-		s = quotify_string (lease -> filename, MDL);
-		if (s) {
-			fprintf (leaseFile, "  server-name \"%s\";\n", s);
+	if (lease->server_name != NULL) {
+		s = quotify_string(lease->server_name, MDL);
+		if (s != NULL) {
+			fprintf(leaseFile, "  server-name \"%s\";\n", s);
 			if (errno) {
 				++errors;
 				errno = 0;
 			}
-			dfree (s, MDL);
+			dfree(s, MDL);
 		} else
 			++errors;
 	}
@@ -2355,49 +2412,27 @@ int write_client_lease (client, lease, rewrite, makesure)
 				      client, write_lease_option);
 	}
 
-	/* Note: the following is not a Y2K bug - it's a Y1.9K bug.   Until
-	   somebody invents a time machine, I think we can safely disregard
-	   it. */
-	t = gmtime (&lease -> renewal);
-	fprintf (leaseFile,
-		 "  renew %d %d/%d/%d %02d:%02d:%02d;\n",
-		 t -> tm_wday, t -> tm_year + 1900,
-		 t -> tm_mon + 1, t -> tm_mday,
-		 t -> tm_hour, t -> tm_min, t -> tm_sec);
-	if (errno != 0) {
+	tval = print_time(lease->renewal);
+	if (tval == NULL ||
+	    fprintf(leaseFile, "  renew %s\n", tval) < 0)
 		errors++;
-		errno = 0;
-	}
-	t = gmtime (&lease -> rebind);
-	fprintf (leaseFile,
-		 "  rebind %d %d/%d/%d %02d:%02d:%02d;\n",
-		 t -> tm_wday, t -> tm_year + 1900,
-		 t -> tm_mon + 1, t -> tm_mday,
-		 t -> tm_hour, t -> tm_min, t -> tm_sec);
-	if (errno != 0) {
+
+	tval = print_time(lease->rebind);
+	if (tval == NULL ||
+	    fprintf(leaseFile, "  rebind %s\n", tval) < 0)
 		errors++;
-		errno = 0;
-	}
-	t = gmtime (&lease -> expiry);
-	fprintf (leaseFile,
-		 "  expire %d %d/%d/%d %02d:%02d:%02d;\n",
-		 t -> tm_wday, t -> tm_year + 1900,
-		 t -> tm_mon + 1, t -> tm_mday,
-		 t -> tm_hour, t -> tm_min, t -> tm_sec);
-	if (errno != 0) {
+
+	tval = print_time(lease->expiry);
+	if (tval == NULL ||
+	    fprintf(leaseFile, "  expire %s\n", tval) < 0)
 		errors++;
-		errno = 0;
-	}
-	fprintf (leaseFile, "}\n");
-	if (errno != 0) {
+
+	if (fprintf(leaseFile, "}\n") < 0)
 		errors++;
-		errno = 0;
-	}
-	fflush (leaseFile);
-	if (errno != 0) {
+
+	if (fflush(leaseFile) != 0)
 		errors++;
-		errno = 0;
-	}
+
 	if (!errors && makesure) {
 		if (fsync (fileno (leaseFile)) < 0) {
 			log_info ("write_client_lease: %m");
@@ -2722,7 +2757,6 @@ void go_daemon ()
 {
 	static int state = 0;
 	int pid;
-	int i;
 
 	/* Don't become a daemon if the user requested otherwise. */
 	if (no_daemon) {
@@ -2752,14 +2786,10 @@ void go_daemon ()
         close(2);
 
 	/* Reopen them on /dev/null. */
-	i = open ("/dev/null", O_RDWR);
-	if (i == 0)
-		i = open ("/dev/null", O_RDWR);
-	if (i == 1) {
-		i = open ("/dev/null", O_RDWR);
-		log_perror = 0; /* No sense logging to /dev/null. */
-	} else if (i != -1)
-		close (i);
+	open("/dev/null", O_RDWR);
+	open("/dev/null", O_RDWR);
+	open("/dev/null", O_RDWR);
+	log_perror = 0; /* No sense logging to /dev/null. */
 
 	write_client_pid_file ();
 }
@@ -3071,7 +3101,7 @@ void client_dns_update_timeout (void *cp)
 		}
 	}
 }
-			
+
 /* See if we should do a DNS update, and if so, do it. */
 
 isc_result_t client_dns_update (struct client_state *client, int addp, int ttl)
@@ -3167,7 +3197,7 @@ isc_result_t client_dns_update (struct client_state *client, int addp, int ttl)
 			rcode = ddns_update_a (&ddns_fwd_name,
 					       client -> active -> address,
 					       &ddns_dhcid, ttl,
-					       1);
+					       1, 1);
 		else
 			rcode = ddns_remove_a (&ddns_fwd_name,
 					       client -> active -> address,
