@@ -3,7 +3,7 @@
    Lexical scanner for dhcpd config file... */
 
 /*
- * Copyright (c) 2004 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2006 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -34,7 +34,7 @@
 
 #ifndef lint
 static char copyright[] =
-"$Id: conflex.c,v 1.92.2.9 2004/11/24 17:39:15 dhankins Exp $ Copyright (c) 2004 Internet Systems Consortium.  All rights reserved.\n";
+"$Id: conflex.c,v 1.105.8.1 2007/05/29 17:49:44 each Exp $ Copyright (c) 2004-2006 Internet Systems Consortium.  All rights reserved.\n";
 #endif /* not lint */
 
 #include "dhcpd.h"
@@ -46,7 +46,7 @@ static void skip_to_eol PROTO ((struct parse *));
 static enum dhcp_token read_string PROTO ((struct parse *));
 static enum dhcp_token read_number PROTO ((int, struct parse *));
 static enum dhcp_token read_num_or_name PROTO ((int, struct parse *));
-static enum dhcp_token intern PROTO ((char *, enum dhcp_token));
+static enum dhcp_token intern PROTO ((unsigned char *, enum dhcp_token));
 
 isc_result_t new_parse (cfile, file, inbuf, buflen, name, eolp)
 	struct parse **cfile;
@@ -95,10 +95,14 @@ isc_result_t new_parse (cfile, file, inbuf, buflen, name, eolp)
 isc_result_t end_parse (cfile)
 	struct parse **cfile;
 {
-	if ((*cfile) -> bufsiz)
-		dfree ((*cfile) -> inbuf, MDL);
-	dfree (*cfile, MDL);
-	*cfile = (struct parse *)0;
+	/* "Memory" config files have no file. */
+	if ((*cfile)->file != -1)
+		close((*cfile)->file);
+
+	if ((*cfile)->bufsiz)
+		dfree((*cfile)->inbuf, MDL);
+	dfree(*cfile, MDL);
+	*cfile = NULL;
 	return ISC_R_SUCCESS;
 }
 
@@ -410,22 +414,62 @@ static enum dhcp_token read_number (c, cfile)
 	int c;
 	struct parse *cfile;
 {
+#ifdef OLD_LEXER
 	int seenx = 0;
+#endif
 	int i = 0;
 	int token = NUMBER;
 
 	cfile -> tokbuf [i++] = c;
 	for (; i < sizeof cfile -> tokbuf; i++) {
 		c = get_char (cfile);
-		if (!seenx && c == 'x') {
-			seenx = 1;
+
 #ifndef OLD_LEXER
-		} else if (isascii (c) && !isxdigit (c) &&
-			   (c == '-' || c == '_' || isalpha (c))) {
-			token = NAME;
-		} else if (isascii (c) && !isdigit (c) && isxdigit (c)) {
-			token = NUMBER_OR_NAME;
-#endif
+		/* Promote NUMBER -> NUMBER_OR_NAME -> NAME, never demote.
+		 * Except in the case of '0x' syntax hex, which gets called
+		 * a NAME at '0x', and returned to NUMBER_OR_NAME once it's
+		 * verified to be at least 0xf or less.
+		 */
+		switch(isascii(c) ? token : BREAK) {
+		    case NUMBER:
+			if(isdigit(c))
+				break;
+			/* FALLTHROUGH */
+		    case NUMBER_OR_NAME:
+			if(isxdigit(c)) {
+				token = NUMBER_OR_NAME;
+				break;
+			}
+			/* FALLTHROUGH */
+		    case NAME:
+			if((i == 2) && isxdigit(c) &&
+				(cfile->tokbuf[0] == '0') &&
+				((cfile->tokbuf[1] == 'x') ||
+				 (cfile->tokbuf[1] == 'X'))) {
+				token = NUMBER_OR_NAME;
+				break;
+			} else if(((c == '-') || (c == '_') || isalnum(c))) {
+				token = NAME;
+				break;
+			}
+			/* FALLTHROUGH */
+		    case BREAK:
+			/* At this point c is either EOF or part of the next
+			 * token.  If not EOF, rewind the file one byte so
+			 * the next token is read from there.
+			 */
+			if(c != EOF) {
+				cfile->bufix--;
+				cfile->ugflag = 1;
+			}
+			goto end_read;
+
+		    default:
+			log_fatal("read_number():%s:%d: impossible case", MDL);
+		}
+#else /* OLD_LEXER */
+		if (!seenx && (c == 'x')) {
+			seenx = 1;
 		} else if (!isascii (c) || !isxdigit (c)) {
 			if (c != EOF) {
 				cfile -> bufix--;
@@ -433,16 +477,22 @@ static enum dhcp_token read_number (c, cfile)
 			}
 			break;
 		}
+#endif /* OLD_LEXER */
+
 		cfile -> tokbuf [i] = c;
 	}
+
 	if (i == sizeof cfile -> tokbuf) {
 		parse_warn (cfile,
 			    "numeric token larger than internal buffer");
 		--i;
 	}
+
+  end_read:
 	cfile -> tokbuf [i] = 0;
 	cfile -> tlen = i;
 	cfile -> tval = cfile -> tokbuf;
+
 	return token;
 }
 
@@ -478,7 +528,7 @@ static enum dhcp_token read_num_or_name (c, cfile)
 }
 
 static enum dhcp_token intern (atom, dfv)
-	char *atom;
+	unsigned char *atom;
 	enum dhcp_token dfv;
 {
 	if (!isascii (atom [0]))
@@ -527,6 +577,8 @@ static enum dhcp_token intern (atom, dfv)
 			return ADDRESS;
 		if (!strcasecmp (atom + 1, "ctive"))
 			return TOKEN_ACTIVE;
+		if (!strcasecmp (atom + 1, "tsfp"))
+			return ATSFP;
 		break;
 	      case 'b':
 		if (!strcasecmp (atom + 1, "ackup"))
@@ -595,14 +647,20 @@ static enum dhcp_token intern (atom, dfv)
 			return CLTT;
 		break;
 	      case 'd':
+		if (!strcasecmp(atom + 1, "b-time-format"))
+			return DB_TIME_FORMAT;
 		if (!strcasecmp (atom + 1, "ns-update"))
 			return DNS_UPDATE;
 		if (!strcasecmp (atom + 1, "ns-delete"))
 			return DNS_DELETE;
 		if (!strcasecmp (atom + 1, "omain"))
 			return DOMAIN;
-		if (!strcasecmp (atom + 1, "omain-name"))
-			return DOMAIN_NAME;
+		if (!strncasecmp (atom + 1, "omain-", 6)) {
+			if (!strcasecmp(atom + 7, "name"))
+				return DOMAIN_NAME;
+			if (!strcasecmp(atom + 7, "list"))
+				return DOMAIN_LIST;
+		}
 		if (!strcasecmp (atom + 1, "o-forward-update"))
 			return DO_FORWARD_UPDATE;
 		if (!strcasecmp (atom + 1, "ebug"))
@@ -659,6 +717,8 @@ static enum dhcp_token intern (atom, dfv)
 		}
 		if (!strcasecmp (atom + 1, "ncode-int"))
 			return ENCODE_INT;
+		if (!strcasecmp(atom + 1, "poch"))
+			return EPOCH;
 		if (!strcasecmp (atom + 1, "thernet"))
 			return ETHERNET;
 		if (!strcasecmp (atom + 1, "nds"))
@@ -676,6 +736,8 @@ static enum dhcp_token intern (atom, dfv)
 			return EVAL;
 		if (!strcasecmp (atom + 1, "ncapsulate"))
 			return ENCAPSULATE;
+		if (!strcasecmp(atom + 1, "xecute"))
+			return EXECUTE;
 		break;
 	      case 'f':
 		if (!strcasecmp (atom + 1, "atal"))
@@ -704,6 +766,8 @@ static enum dhcp_token intern (atom, dfv)
 			return GET_LEASE_HOSTNAMES;
 		break;
 	      case 'h':
+		if (!strcasecmp(atom + 1, "ash"))
+			return HASH;
 		if (!strcasecmp (atom + 1, "ba"))
 			return HBA;
 		if (!strcasecmp (atom + 1, "ost"))
@@ -753,18 +817,26 @@ static enum dhcp_token intern (atom, dfv)
 			return KEY;
 		break;
 	      case 'l':
+		if (!strcasecmp (atom + 1, "case"))
+			return LCASE;
 		if (!strcasecmp (atom + 1, "ease"))
 			return LEASE;
 		if (!strcasecmp (atom + 1, "eased-address"))
 			return LEASED_ADDRESS;
 		if (!strcasecmp (atom + 1, "ease-time"))
 			return LEASE_TIME;
+		if (!strcasecmp(atom + 1, "easequery"))
+			return LEASEQUERY;
+		if (!strcasecmp(atom + 1, "ength"))
+			return LENGTH;
 		if (!strcasecmp (atom + 1, "imit"))
 			return LIMIT;
 		if (!strcasecmp (atom + 1, "et"))
 			return LET;
 		if (!strcasecmp (atom + 1, "oad"))
 			return LOAD;
+		if (!strcasecmp(atom + 1, "ocal"))
+			return LOCAL;
 		if (!strcasecmp (atom + 1, "og"))
 			return LOG;
 		break;
@@ -772,8 +844,16 @@ static enum dhcp_token intern (atom, dfv)
 		if (!strncasecmp (atom + 1, "ax", 2)) {
 			if (!atom [3])
 				return TOKEN_MAX;
-			if (!strcasecmp (atom + 3, "-lease-time"))
-				return MAX_LEASE_TIME;
+			if (!strcasecmp (atom + 3, "-balance"))
+				return MAX_BALANCE;
+			if (!strncasecmp (atom + 3, "-lease-", 7)) {
+				if (!strcasecmp(atom + 10, "misbalance"))
+					return MAX_LEASE_MISBALANCE;
+				if (!strcasecmp(atom + 10, "ownership"))
+					return MAX_LEASE_OWNERSHIP;
+				if (!strcasecmp(atom + 10, "time"))
+					return MAX_LEASE_TIME;
+			}
 			if (!strcasecmp (atom + 3, "-transmit-idle"))
 				return MAX_TRANSMIT_IDLE;
 			if (!strcasecmp (atom + 3, "-response-delay"))
@@ -782,6 +862,8 @@ static enum dhcp_token intern (atom, dfv)
 				return MAX_UNACKED_UPDATES;
 		}
 		if (!strncasecmp (atom + 1, "in-", 3)) {
+			if (!strcasecmp (atom + 4, "balance"))
+				return MIN_BALANCE;
 			if (!strcasecmp (atom + 4, "lease-time"))
 				return MIN_LEASE_TIME;
 			if (!strcasecmp (atom + 4, "secs"))
@@ -936,71 +1018,122 @@ static enum dhcp_token intern (atom, dfv)
 			return REFRESH;
 		break;
 	      case 's':
-		if (!strcasecmp (atom + 1, "tate"))
-			return STATE;
-		if (!strcasecmp (atom + 1, "ecret"))
-			return SECRET;
-		if (!strcasecmp (atom + 1, "ervfail"))
-			return NS_SERVFAIL;
-		if (!strcasecmp (atom + 1, "witch"))
-			return SWITCH;
-		if (!strcasecmp (atom + 1, "igned"))
-			return SIGNED;
-		if (!strcasecmp (atom + 1, "tring"))
-			return STRING_TOKEN;
-		if (!strcasecmp (atom + 1, "uffix"))
-			return SUFFIX;
-		if (!strcasecmp (atom + 1, "earch"))
-			return SEARCH;
-		if (!strcasecmp (atom + 1, "tarts"))
-			return STARTS;
-		if (!strcasecmp (atom + 1, "iaddr"))
-			return SIADDR;
-		if (!strcasecmp (atom + 1, "hared-network"))
-			return SHARED_NETWORK;
-		if (!strcasecmp (atom + 1, "econdary"))
-			return SECONDARY;
-		if (!strcasecmp (atom + 1, "erver-name"))
-			return SERVER_NAME;
-		if (!strcasecmp (atom + 1, "erver-identifier"))
-			return SERVER_IDENTIFIER;
-		if (!strcasecmp (atom + 1, "erver"))
-			return SERVER;
-		if (!strcasecmp (atom + 1, "elect-timeout"))
-			return SELECT_TIMEOUT;
-		if (!strcasecmp (atom + 1, "elect"))
-			return SELECT;
-		if (!strcasecmp (atom + 1, "end"))
-			return SEND;
-		if (!strcasecmp (atom + 1, "cript"))
-			return SCRIPT;
-		if (!strcasecmp (atom + 1, "upersede"))
-			return SUPERSEDE;
-		if (!strncasecmp (atom + 1, "ub", 2)) {
-			if (!strcasecmp (atom + 3, "string"))
-				return SUBSTRING;
-			if (!strcasecmp (atom + 3, "net"))
-				return SUBNET;
-			if (!strcasecmp (atom + 3, "class"))
-				return SUBCLASS;
+                if (!strcasecmp(atom + 1, "cript"))
+                        return SCRIPT;
+		if (isascii(atom[1]) && tolower(atom[1]) == 'e') {
+                        if (!strcasecmp(atom + 2, "arch"))
+                                return SEARCH;
+			if (isascii(atom[2]) && tolower(atom[2]) == 'c') {
+				if (!strncasecmp(atom + 3, "ond", 3)) {
+                                        if (!strcasecmp(atom + 6, "ary"))
+                                                return SECONDARY;
+                                        if (!strcasecmp(atom + 6, "s"))
+                                                return SECONDS;
+					break;
+				}
+                                if (!strcasecmp(atom + 3, "ret"))
+                                        return SECRET;
+				break;
+			}
+			if (!strncasecmp(atom + 2, "lect", 4)) {
+                                if (atom[6] == '\0')
+                                        return SELECT;
+                                if (!strcasecmp(atom + 6, "-timeout"))
+                                        return SELECT_TIMEOUT;
+				break;
+			}
+                        if (!strcasecmp(atom + 2, "nd"))
+                                return SEND;
+			if (!strncasecmp(atom + 2, "rv", 2)) {
+				if (!strncasecmp(atom + 4, "er", 2)) {
+                                        if (atom[6] == '\0')
+                                                return SERVER;
+					if (atom[6] == '-') {
+                                                if (!strcasecmp(atom + 7,
+								"name"))
+                                                        return SERVER_NAME;
+                                                if (!strcasecmp(atom + 7,
+								"identifier"))
+                                                      return SERVER_IDENTIFIER;
+						break;
+					}
+					break;
+				}
+                                if (!strcasecmp(atom + 4, "fail"))
+                                        return NS_SERVFAIL;
+				break;
+			}
+                        if (!strcasecmp(atom + 2, "t"))
+                                return TOKEN_SET;
 			break;
 		}
-		if (!strcasecmp (atom + 1, "pawn"))
-			return SPAWN;
-		if (!strcasecmp (atom + 1, "pace"))
-			return SPACE;
-		if (!strcasecmp (atom + 1, "tatic"))
-			return STATIC;
-		if (!strcasecmp (atom + 1, "plit"))
-			return SPLIT;
-		if (!strcasecmp (atom + 1, "et"))
-			return TOKEN_SET;
-		if (!strcasecmp (atom + 1, "econds"))
-			return SECONDS;
-		if (!strcasecmp (atom + 1, "hutdown"))
-			return SHUTDOWN;
-		if (!strcasecmp (atom + 1, "tartup"))
-			return STARTUP;
+		if (isascii(atom[1]) && tolower(atom[1]) == 'h') {
+                        if (!strcasecmp(atom + 2, "ared-network"))
+                                return SHARED_NETWORK;
+                        if (!strcasecmp(atom + 2, "utdown"))
+                                return SHUTDOWN;
+			break;
+		}
+		if (isascii(atom[1]) && tolower(atom[1]) == 'i') {
+                        if (!strcasecmp(atom + 2, "addr"))
+                                return SIADDR;
+                        if (!strcasecmp(atom + 2, "gned"))
+                                return SIGNED;
+                        if (!strcasecmp(atom + 2, "ze"))
+                                return SIZE;
+			break;
+		}
+		if (isascii(atom[1]) && tolower(atom[1]) == 'p') {
+			if (isascii(atom[2]) && tolower(atom[2]) == 'a') {
+                                if (!strcasecmp(atom + 3, "ce"))
+                                        return SPACE;
+                                if (!strcasecmp(atom + 3, "wn"))
+                                        return SPAWN;
+				break;
+			}
+                        if (!strcasecmp(atom + 2, "lit"))
+                                return SPLIT;
+			break;
+		}
+		if (isascii(atom[1]) && tolower(atom[1]) == 't') {
+			if (isascii(atom[2]) && tolower(atom[2]) == 'a') {
+				if(!strncasecmp(atom + 3, "rt", 2)) {
+                                         if (!strcasecmp(atom + 5, "s"))
+                                                 return STARTS;
+                                         if (!strcasecmp(atom + 5, "up"))
+                                                 return STARTUP;
+					break;
+				}
+				if (isascii(atom[3]) &&
+				    tolower(atom[3]) == 't') {
+                                        if (!strcasecmp(atom + 4, "e"))
+                                                return STATE;
+                                        if (!strcasecmp(atom + 4, "ic"))
+                                                return STATIC;
+					break;
+				}
+			}
+                        if (!strcasecmp(atom + 2, "ring"))
+                                return STRING_TOKEN;
+			break;
+		}
+                if (!strncasecmp(atom + 1, "ub", 2)) {
+                        if (!strcasecmp(atom + 3, "class"))
+                                return SUBCLASS;
+                        if (!strcasecmp(atom + 3, "net"))
+                                return SUBNET;
+                        if (!strcasecmp(atom + 3, "string"))
+                                return SUBSTRING;
+                        break;
+                }
+		if (isascii(atom[1]) && tolower(atom[1]) == 'u') {
+                        if (!strcasecmp(atom + 2, "ffix"))
+                                return SUFFIX;
+                        if (!strcasecmp(atom + 2, "persede"))
+                                return SUPERSEDE;
+		}
+                if (!strcasecmp(atom + 1, "witch"))
+                        return SWITCH;
 		break;
 	      case 't':
 		if (!strcasecmp (atom + 1, "imestamp"))
@@ -1019,6 +1152,8 @@ static enum dhcp_token intern (atom, dfv)
 			return TRANSMISSION;
 		break;
 	      case 'u':
+		if (!strcasecmp (atom + 1, "case"))
+			return UCASE;
 		if (!strcasecmp (atom + 1, "nset"))
 			return UNSET;
 		if (!strcasecmp (atom + 1, "nsigned"))
@@ -1060,6 +1195,8 @@ static enum dhcp_token intern (atom, dfv)
 	      case 'w':
 		if (!strcasecmp (atom + 1, "ith"))
 			return WITH;
+		if (!strcasecmp(atom + 1, "idth"))
+			return WIDTH;
 		break;
 	      case 'y':
 		if (!strcasecmp (atom + 1, "iaddr"))
