@@ -3,7 +3,7 @@
    Definitions for address trees... */
 
 /*
- * Copyright (c) 2004 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2007 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -190,13 +190,15 @@ enum expr_op {
 	expr_binary_and,
 	expr_binary_or,
 	expr_binary_xor,
-	expr_client_state
+	expr_client_state,
+	expr_ucase,
+	expr_lcase
 };
 
 struct expression {
 	int refcnt;
 	enum expr_op op;
-	union {
+	union expr_union {
 		struct {
 			struct expression *expr;
 			struct expression *offset;
@@ -216,6 +218,8 @@ struct expression {
 			struct expression *expr;
 			struct expression *len;
 		} suffix;
+		struct expression *lcase;
+		struct expression *ucase;
 		struct option *option;
 		struct option *config_option;
 		struct {
@@ -233,7 +237,7 @@ struct expression {
 		struct {
 			struct expression *base;
 			struct expression *width;
-			struct expression *seperator;
+			struct expression *separator;
 			struct expression *buffer;
 		} b2a;
 		struct {
@@ -320,11 +324,14 @@ struct universe {
 			    struct option_state *, struct option_state *,
 			    struct binding_scope **,
 			    struct universe *);
+	u_int32_t (*get_tag) (const unsigned char *);
 	void (*store_tag) PROTO ((unsigned char *, u_int32_t));
+	u_int32_t (*get_length) (const unsigned char *);
 	void (*store_length) PROTO ((unsigned char *, u_int32_t));
 	int tag_size, length_size;
-	option_hash_t *hash;
-	struct option *options [256];
+	unsigned end;
+	option_name_hash_t *name_hash;
+	option_code_hash_t *code_hash;
 	struct option *enc_opt;
 	int index;
 };
@@ -334,4 +341,5 @@ struct option {
 	const char *format;
 	struct universe *universe;
 	unsigned code;
+	int refcnt;
 };
