@@ -3,7 +3,7 @@
    DHCP Server Daemon. */
 
 /*
- * Copyright (c) 2004-2008 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2007 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1996-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -34,11 +34,11 @@
 
 #ifndef lint
 static char ocopyright[] =
-"$Id: dhcpd.c,v 1.115.2.23 2008/05/14 20:54:24 dhankins Exp $ Copyright 2004-2008 Internet Systems Consortium.";
+"$Id: dhcpd.c,v 1.121.42.2 2007/05/29 17:49:44 each Exp $ Copyright 2004-2007 Internet Systems Consortium.";
 #endif
 
   static char copyright[] =
-"Copyright 2004-2008 Internet Systems Consortium.";
+"Copyright 2004-2007 Internet Systems Consortium.";
 static char arr [] = "All rights reserved.";
 static char message [] = "Internet Systems Consortium DHCP Server";
 static char url [] = "For info, please visit http://www.isc.org/sw/dhcp/";
@@ -197,12 +197,14 @@ int main (argc, argv, envp)
 	int argc;
 	char **argv, **envp;
 {
+	int fd;
 	int i, status;
 	struct servent *ent;
-	char *s;
+	unsigned char *s;
 	int cftest = 0;
 	int lftest = 0;
 #ifndef DEBUG
+	int pidfilewritten = 0;
 	int pid;
 	char pbuf [20];
 	int daemon = 1;
@@ -225,15 +227,18 @@ int main (argc, argv, envp)
 	char *traceoutfile = (char *)0;
 #endif
 
-	/* Make sure we have stdin, stdout and stderr. */
-	status = open ("/dev/null", O_RDWR);
-	if (status == 0)
-		status = open ("/dev/null", O_RDWR);
-	if (status == 1) {
-		status = open ("/dev/null", O_RDWR);
-		log_perror = 0; /* No sense logging to /dev/null. */
-	} else if (status != -1)
-		close (status);
+        /* Make sure that file descriptors 0 (stdin), 1, (stdout), and
+           2 (stderr) are open. To do this, we assume that when we
+           open a file the lowest available file decriptor is used. */
+        fd = open("/dev/null", O_RDWR);
+        if (fd == 0)
+                fd = open("/dev/null", O_RDWR);
+        if (fd == 1)
+                fd = open("/dev/null", O_RDWR);
+        if (fd == 2)
+                log_perror = 0; /* No sense logging to /dev/null. */
+        else if (fd != -1)
+                close(fd);
 
 	/* Set up the client classification system. */
 	classification_setup ();
@@ -537,32 +542,28 @@ int main (argc, argv, envp)
 
 	/* Read previous pid file. */
 	if ((i = open (path_dhcpd_pid, O_RDONLY)) >= 0) {
-		status = read(i, pbuf, (sizeof pbuf) - 1);
+		status = read (i, pbuf, (sizeof pbuf) - 1);
 		close (i);
 		if (status > 0) {
-			pbuf[status] = 0;
-			pid = atoi(pbuf);
+			pbuf [status] = 0;
+			pid = atoi (pbuf);
 
-			/*
-                         * If there was a previous server process and it's
-                         * is still running, abort
-                         */
-			if (!pid || (pid != getpid() && kill(pid, 0) == 0))
-				log_fatal("There's already a "
-                                          "DHCP server running.");
+			/* If the previous server process is not still running,
+			   write a new pid file immediately. */
+			if (pid && (pid == getpid() || kill (pid, 0) < 0)) {
+				unlink (path_dhcpd_pid);
+				if ((i = open (path_dhcpd_pid,
+					O_WRONLY | O_CREAT, 0644)) >= 0) {
+				    sprintf (pbuf, "%d\n", (int)getpid ());
+				    write (i, pbuf, strlen (pbuf));
+				    close (i);
+				    pidfilewritten = 1;
+				}
+			} else
+				log_fatal ("There's already a DHCP server running.");
 		}
 	}
 
-        /* Write new pid file. */
-        if ((i = open(path_dhcpd_pid, O_WRONLY|O_CREAT|O_TRUNC, 0644)) >= 0) {
-                sprintf(pbuf, "%d\n", (int) getpid());
-                write(i, pbuf, strlen(pbuf));
-                close(i);
-        } else {
-                log_error("Can't create PID file %s: %m.", path_dhcpd_pid);
-        }
-
-
 	/* If we were requested to log to stdout on the command line,
 	   keep doing so; otherwise, stop. */
 	if (log_perror == -1)
@@ -572,10 +573,33 @@ int main (argc, argv, envp)
 
 	if (daemon) {
 		/* Become session leader and get pid... */
-		close (0);
-		close (1);
-		close (2);
-		pid = setsid ();
+		pid = setsid();
+
+                /* Close standard I/O descriptors. */
+                close(0);
+                close(1);
+                close(2);
+
+                /* Reopen them on /dev/null. */
+                open("/dev/null", O_RDWR);
+                open("/dev/null", O_RDWR);
+                open("/dev/null", O_RDWR);
+                log_perror = 0; /* No sense logging to /dev/null. */
+	}
+
+	/* If we didn't write the pid file earlier because we found a
+	   process running the logged pid, but we made it to here,
+	   meaning nothing is listening on the bootp port, then write
+	   the pid file out - what's in it now is bogus anyway. */
+	if (!pidfilewritten) {
+		unlink (path_dhcpd_pid);
+		if ((i = open (path_dhcpd_pid,
+			       O_WRONLY | O_CREAT, 0644)) >= 0) {
+			sprintf (pbuf, "%d\n", (int)getpid ());
+			write (i, pbuf, strlen (pbuf));
+			close (i);
+			pidfilewritten = 1;
+		}
 	}
 #endif /* !DEBUG */
 
@@ -586,6 +610,10 @@ int main (argc, argv, envp)
 	dmalloc_outstanding = 0;
 #endif
 
+#if defined (DEBUG_RC_HISTORY_EXHAUSTIVELY)
+	dump_rc_history ();
+#endif
+
 	omapi_set_int_value ((omapi_object_t *)dhcp_control_object,
 			     (omapi_object_t *)0, "state", server_running);
 
